"""The module that contains the CLI commands.

Classes:
    Command: The commands that can be executed by the CLI.
    Generate: The class that executes the commands based on the provided arguments.
"""

import datetime
import enum
import importlib
import pathlib

import cookiecutter.main  # type: ignore
import typer
import yaml

from .commons import (
    DOCS_PATH,
    ROOT_PATH,
    TEMPLATES_PATH,
    BaseDocsGenerator,
    get_imports_from_module,
    read_pyproject,
    recursive_writer,
    to_snake_case,
)


class Generate:
    """The class that executes the options based on the provided arguments.

    Methods:
        execute: Executes the option based on the provided arguments.
    """

    class GenerateOptionsEnum(str, enum.Enum):
        """The options that can be executed by the CLI.

        Options:
            docs
            docstring
            index-file
            installation
            licence
            logo
            mkdocs
            project
            readme
            requirements
            usage
        """

        docs = "docs"
        docstring = "docstring"
        index_file = "index-file"
        installation = "installation"
        licence = "licence"
        logo = "logo"
        mkdocs = "mkdocs"
        project = "project"
        readme = "readme"
        requirements = "requirements"
        usage = "usage"

    @staticmethod
    def execute(option: GenerateOptionsEnum, project_name: str | None = None) -> None:
        """Executes the option based on the provided arguments."""
        match option:
            case Generate.GenerateOptionsEnum.docs:
                Generate.execute(Generate.GenerateOptionsEnum.licence)

                Generate.execute(Generate.GenerateOptionsEnum.index_file)
                Generate.execute(Generate.GenerateOptionsEnum.requirements)
                Generate.execute(Generate.GenerateOptionsEnum.installation)
                Generate.execute(Generate.GenerateOptionsEnum.usage)
                Generate.execute(Generate.GenerateOptionsEnum.readme)

                Generate.execute(Generate.GenerateOptionsEnum.docstring)
                Generate.execute(Generate.GenerateOptionsEnum.mkdocs)

            case Generate.GenerateOptionsEnum.docstring:
                project_slug = to_snake_case(str(ROOT_PATH.absolute()).split("/")[-1])
                typer.echo(f"Generating docs for {project_slug}")

                project_module = importlib.import_module(project_slug)
                heading_level = 3
                imports_from_module = get_imports_from_module(project_module, heading_level)

                typer.echo("Clearing API Reference folder")
                api_path = DOCS_PATH / "api"
                for doc in api_path.glob("*"):
                    doc.unlink()

                for doc in imports_from_module:
                    recursive_writer(doc)

                typer.echo(f"Generated docs for {project_slug}")

            case Generate.GenerateOptionsEnum.index_file:
                pyproject_content = read_pyproject()
                docs_generator = BaseDocsGenerator(pyproject_content, project_name=project_name)

                docs_generator.render("index.md")

            case Generate.GenerateOptionsEnum.installation:
                pyproject_content = read_pyproject()
                docs_generator = BaseDocsGenerator(pyproject_content, project_name=project_name)

                docs_generator.render("installation.md")

            case Generate.GenerateOptionsEnum.licence:
                pyproject_content = read_pyproject()
                docs_generator = BaseDocsGenerator(pyproject_content, project_name=project_name)

                year = datetime.datetime.now(datetime.timezone.utc).year
                docs_generator.render("license.md", year=year)
                docs_generator.render("license.md", ROOT_PATH, "LICENSE", year=year)

            case Generate.GenerateOptionsEnum.logo:
                pyproject_content = read_pyproject()
                docs_generator = BaseDocsGenerator(pyproject_content, project_name=project_name)

                assert project_name is not None, "Project name is required for project"
                path = DOCS_PATH / "images"
                docs_generator.render("logo.svg", path, f"{project_name}.svg")

            case Generate.GenerateOptionsEnum.mkdocs:
                pyproject_content = read_pyproject()
                docs_generator = BaseDocsGenerator(pyproject_content, project_name=project_name)

                Generate.execute(Generate.GenerateOptionsEnum.docstring)

                # prerender mkdocs.yml
                docs_generator.render("mkdocs.yml", ROOT_PATH / "docs")

                # append API references to mkdocs.yml
                typer.echo("Generating API Reference")

                typer.echo("Reading temp mkdocs")
                with pathlib.Path.open(ROOT_PATH / "docs" / "mkdocs.yml") as mkdocs_file:
                    mkdocs_dict = yaml.safe_load(mkdocs_file.read())

                typer.echo("Getting references")
                api_references = sorted([i.name for i in pathlib.Path.iterdir(DOCS_PATH / "api")])

                typer.echo("Rendering mkdocs_dict with the data")
                for section in mkdocs_dict["nav"]:
                    if "API Reference" in section:
                        section["API Reference"] = [
                            {"_".join(i.split("_")[1:]).rsplit(".", 1)[0]: f"api/{i}"} for i in api_references
                        ]

                typer.echo("Writing mkdocs")
                with pathlib.Path.open(ROOT_PATH / "docs" / "mkdocs.yml", "w") as mkdocs_file:
                    yaml.dump(mkdocs_dict, mkdocs_file)

                typer.echo("Generated API Reference")

            case Generate.GenerateOptionsEnum.project:
                assert project_name is not None, "Project name is required for project"

                typer.echo(f"Generating project with name: {project_name}")

                typer.echo("Cloning template")
                cookiecutter.main.cookiecutter(  # type: ignore
                    str(TEMPLATES_PATH / "cookiecutter"),
                    no_input=True,
                    extra_context={
                        "project_name": project_name,
                        "project_slug": to_snake_case(project_name),
                        "year": datetime.datetime.now(datetime.timezone.utc).year,
                    },
                    overwrite_if_exists=True,
                )

                Generate.execute(Generate.GenerateOptionsEnum.logo, project_name)
                Generate.execute(Generate.GenerateOptionsEnum.licence)

            case Generate.GenerateOptionsEnum.readme:
                pyproject_content = read_pyproject()
                docs_generator = BaseDocsGenerator(pyproject_content, project_name=project_name)

                Generate.execute(Generate.GenerateOptionsEnum.index_file)
                Generate.execute(Generate.GenerateOptionsEnum.requirements)
                Generate.execute(Generate.GenerateOptionsEnum.installation)
                Generate.execute(Generate.GenerateOptionsEnum.usage)

                docs_generator.concatenate("README.md", ROOT_PATH)

            case Generate.GenerateOptionsEnum.requirements:
                pyproject_content = read_pyproject()
                docs_generator = BaseDocsGenerator(pyproject_content, project_name=project_name)

                docs_generator.render("requirements.md")

            case Generate.GenerateOptionsEnum.usage:
                pyproject_content = read_pyproject()
                docs_generator = BaseDocsGenerator(pyproject_content, project_name=project_name)

                docs_generator.render("usage.md")



"""Sthali Core package CLI and main application entry points.

Provides Typer CLI commands for generating, updating, and serving documentation and projects.
"""

import mkdocs.commands.serve
import typer

from .scripts import Generate, commons
from .utils import base, enum_clients, run_server

__all__ = [
    "base",
    "enum_clients",
    "run_server",
]


app = typer.Typer()
state = {}


@app.callback()
def callback() -> None:
    pass


@app.command()
def deploy() -> None:
    pass


@app.command()
def generate(option: Generate.GenerateOptionsEnum, project_name: str | None = None) -> None:
    Generate.execute(option, project_name)


@app.command()
def serve() -> None:
    typer.echo("Serving documentation...")

    config_file_path = str(commons.ROOT_PATH / "docs" / "mkdocs.yml")
    mkdocs.commands.serve.serve(config_file_path)  # type: ignore



"""Dynamically creates an Enum for available clients."""

import enum
import importlib
import pathlib
import types

all = [
    "Clients",
]


class Clients:
    """Discovers and provides an Enum for client modules."""

    def __init__(self, parent_path: pathlib.Path, package: str | None = None) -> None:
        """Initialize Clients.

        Args:
            parent_path: The parent directory path where the 'clients' subdirectory resides.
            package: The package name to use for relative imports. Defaults to the parent_path's name.
        """
        self.clients_directory = parent_path / "clients"
        self.package = package or parent_path.name

    @property
    def _clients_file_path(self) -> list[pathlib.Path]:
        return list(filter(lambda file: file.stem != "__init__", self.clients_directory.glob("*.py")))

    @property
    def _src_path(self) -> pathlib.Path:
        return next(filter(lambda x: x.name in ["src", "site-packages"], self.clients_directory.parents))

    @property
    def clients_map(self) -> dict[str, types.ModuleType]:
        """Return a map of client names to their imported modules.

        Returns:
            A dictionary where keys are client names (derived from filenames)
            and values are the corresponding imported client modules.
        """
        return {
            file_path.stem: importlib.import_module(self._relative_to_formatted(file_path), package=self.package)
            for file_path in self._clients_file_path
        }

    @property
    def enum(self) -> type[enum.Enum]:
        """Dynamically create an Enum representing available clients.

        The Enum members will have names and values corresponding to the client names.

        Returns:
            An Enum type where members are the discovered client names.
        """
        client_enum_map = {key: key for key in self.clients_map}
        return enum.Enum("ClientEnum", client_enum_map)

    def _relative_to_formatted(self, path: pathlib.Path) -> str:
        """Return the relative path to the clients directory."""
        return str(path.relative_to(self._src_path)).strip(".py").replace("/", ".")



"""Common utility functions and classes used across the CLI scripts.

Variables:
    ROOT_PATH (pathlib.Path): The root path of the project.
    DOCS_PATH (pathlib.Path): The path to the docs directory.
    TEMPLATES_PATH (pathlib.Path): The path to the templates directory.
    TEMPLATES (fastapi.templating.Jinja2Templates): Jinja2 templates for rendering documentation.

Functions:
    to_snake_case(string: str) -> str: Converts a given string to snake case.
    read_pyproject(path: pathlib.Path | None) -> dict[str, typing.Any]: Reads the pyproject.toml file and returns its
        content as a dictionary.
    get_imports_from_module(module: types.ModuleType, level: int) -> list[Doc]: Recursively extract documentation
        metadata from a module's __all__ exports.
    recursive_writer(doc: Doc) -> str: Recursively write documentation for a Doc object and its sub-objects.
"""

import pathlib
import re
import types
import typing

import fastapi.templating
import pydantic
import tomli
import typer

ROOT_PATH = pathlib.Path()
DOCS_PATH = ROOT_PATH / "docs" / "docs"
TEMPLATES_PATH = pathlib.Path(__file__).parent / "templates"
TEMPLATES = fastapi.templating.Jinja2Templates(TEMPLATES_PATH)


def read_pyproject(path: pathlib.Path | None = None) -> dict[str, typing.Any]:
    """Reads the pyproject.toml file and returns its content as a dictionary.

    Args:
        path (pathlib.Path): The path to the pyproject.toml file. Defaults to None.

    Returns:
        dict[str, typing.Any]: The content of the pyproject.toml file as a dictionary
    """
    path = ROOT_PATH / "pyproject.toml" or path
    with pathlib.Path.open(path) as pyproject_file:
        return tomli.loads(pyproject_file.read())


def to_snake_case(string: str) -> str:
    """Converts a given string to snake case.

    Args:
        string (str): The string to be converted.

    Returns:
        str: The converted string in snake case.
    """
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", string)
    s2 = re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1)
    return re.sub(r"-", "_", s2).lower()


@pydantic.dataclasses.dataclass
class Doc:
    """Represents documentation metadata for an object.

    Attributes:
        name (str): The name of the object.
        path (pathlib.Path): The output file path for the documentation.
        docstring (str): The docstring or description.
        level (int): The Markdown heading level.
        sub (list["Doc"]): List of Doc objects for sub-objects.
    """

    name: str
    path: pathlib.Path
    docstring: str
    level: int
    sub: list["Doc"]


def recursive_writer(doc: Doc) -> str:
    """Recursively write documentation for a Doc object and its sub-objects.

    Args:
        doc (Doc): The Doc object to write.

    Returns:
        str: The rendered Markdown string for this Doc and its children.
    """

    def _render(name: str, docstring: str, level: int = 3, sub: str | None = None) -> str:
        template = TEMPLATES.get_template("docstring.md")  # type: ignore
        return template.render(  # type: ignore
            name=name,
            docstring=docstring,
            heading_level=level * "#",
            sub=sub,
        )

    typer.echo(f"Writing docstring from import: {doc.name}")

    to_render = _render(doc.name, doc.docstring, doc.level, "\n".join([recursive_writer(s) for s in doc.sub]))
    with pathlib.Path.open(doc.path, "w") as doc_file:
        doc_file.write(to_render)

    return to_render


TypesMapping = typing.Literal["module", "function", "class"]


def get_imports_from_module(module: types.ModuleType, level: int) -> list[Doc]:
    """Recursively extract documentation metadata from a module's __all__ exports.

    Args:
        module (types.ModuleType): The module to inspect.
        level (int): The Markdown heading level.

    Returns:
        list[Doc]: A list of Doc objects representing the module's exports.
    """

    def _get_import_type_mapping(_import: typing.Any) -> TypesMapping | None:
        import_type: typing.Any = type(_import)  # type: ignore
        match import_type:
            case types.ModuleType:
                return "module"
            case types.FunctionType:
                return "function"
            case _ if isinstance(_import, type):
                return "class"
            case _:
                return None

    def _get_docstring(obj: typing.Any) -> str:
        def _get_metadata(obj: typing.Any) -> str | None:
            metadata: list[typing.Any] = getattr(obj, "__metadata__", [])
            field_info = next(filter(lambda x: isinstance(x, pydantic.fields.FieldInfo), metadata), None)
            if field_info:
                return field_info.description
            return None

        def _get_doc(obj: typing.Any) -> str | None:
            return getattr(obj, "__doc__", None)

        metadata = _get_metadata(obj)
        doc = _get_doc(obj)
        return metadata or doc or ""

    typer.echo(f"Getting imports from {module.__name__} module")

    docs: list[Doc] = []
    for i in getattr(module, "__all__", []):
        typer.echo(f"Getting metadata from {i} import")

        _import = getattr(module, i)
        name = _import.__name__
        filename = _import.__name__
        sub = []
        import_type = _get_import_type_mapping(_import) or ""
        api_path = DOCS_PATH / "api"
        if import_type == "module":
            name = _import.__spec__.name
            sub = get_imports_from_module(_import, level + 1)
            filename = to_snake_case(i)

        docs.append(
            Doc(
                name=name,
                path=api_path / f"{import_type}_{filename}.md",
                docstring=_get_docstring(_import),
                level=level,
                sub=sub,
            ),
        )
    return docs


class BaseDocsGenerator:
    """Base class for documentation generators."""

    def __init__(
        self,
        pyproject_content: dict[str, typing.Any],
        organization_name: str | None = None,
        project_name: str | None = None,
    ) -> None:
        """Initialize the documentation generator.

        Args:
            pyproject_content (dict[str, typing.Any]): The content of pyproject.toml.
            organization_name (str | None): The name of the organization. Defaults to None.
            project_name (str | None): The name of the project. Defaults to None.
        """
        self.pyproject_content = pyproject_content
        self.organization_name = organization_name or "project-sthali"
        self.project_name = project_name

    def concatenate(self, file: str, path: pathlib.Path | None = None) -> None:
        """Generate the README file by concatenating documentation files.

        Args:
            file (str): The name of the file to generate (e.g., "index.md").
            path (pathlib.Path | None): The path where the file will be saved. Defaults to DOCS_PATH.
        """
        typer.echo(f"Generating {file}")

        typer.echo(f"Writing {file}")
        path = path or DOCS_PATH
        full_path = path / file
        with pathlib.Path.open(full_path, "w") as f:
            for file_to_concatenate in [
                DOCS_PATH / f"{i}.md" for i in ["index", "requirements", "installation", "usage"]
            ]:
                typer.echo(f"Concatenating doc: {file_to_concatenate.name}")
                with pathlib.Path.open(file_to_concatenate) as _f:
                    f.write(_f.read())
                    f.write("\n")

        typer.echo(f"Generated {full_path}")

    def render(
        self,
        file: str,
        path: pathlib.Path | None = None,
        new_file: str | None = None,
        **kwargs: typing.Any,
    ) -> None:
        """Generate documentation file.

        This method renders a template file with the provided data and writes it to the specified path.

        Args:
            file (str): The name of the file to generate (e.g., "index.md").
            path (pathlib.Path | None): The path where the file will be saved. Defaults to DOCS_PATH.
            new_file (str | None): If provided, the content will be written to this new file instead of `file`.
            **kwargs (typing.Any): Additional keyword arguments to pass to the template rendering.
        """
        typer.echo(f"Generating {file}")

        typer.echo("Rendering the template with the data")
        template = TEMPLATES.get_template(file)  # type: ignore
        content: str = template.render(  # type: ignore
            **self.pyproject_content,
            organization_name=self.organization_name,
            project_name=self.project_name,
            **kwargs or {},
        )

        file = new_file or file
        path = path or DOCS_PATH
        full_path = path / file
        typer.echo(f"Writing {full_path}")
        with pathlib.Path.open(full_path, "w") as f:
            f.write(content)  # type: ignore

        typer.echo(f"Generated {full_path}")



"""Utilities for running the FastAPI server."""

import os

import fastapi
import uvicorn

from .base import App, AppSpecification, Config

all = [
    "run",
    "main",
]


def run(config: type[Config], app_spec: type[AppSpecification], app: type[App]) -> fastapi.FastAPI:
    """Initialize and return the FastAPI application.

    Args:
        config: The configuration class.
        app_spec: The application specification class.
        app: The application class.

    Returns:
        The configured FastAPI application instance.
    """
    app_specification_file_path = os.getenv("APP_SPECIFICATION_FILE_PATH") or "volume/app_specification_sample.json"
    client_config = config(app_specification_file_path)
    client_app = app(app_spec(**client_config.app_specification))
    return client_app.app


def main(app: fastapi.FastAPI, port: int = 8000) -> None:
    """Run the FastAPI application using Uvicorn.

    Args:
        app: The FastAPI application instance to run.
        port: The port number to run the server on. Defaults to 8000.
    """
    uvicorn.run(app, host="0.0.0.0", port=port)



"""Base utilities for Sthali Core."""

import typing

import fastapi
import pydantic

__all__ = [
    "App",
    "AppSpecification",
    "Config",
]


class Config:
    """Manages application configuration."""

    app_specification: dict[str, typing.Any]

    def __init__(self, app_specification_file_path: str) -> None:
        """Initialize Config.

        Args:
            app_specification_file_path: Path to the application specification file.
        """


@pydantic.dataclasses.dataclass
class AppSpecification:
    """Represents the application specification."""


class App:
    """Represents the FastAPI application."""

    app: fastapi.FastAPI

    def __init__(self, app_specification: AppSpecification) -> None:
        """Initialize App.

        Args:
            app_specification: The application specification.
        """
